/*
 *  ========================================================================
 *  DIScrete event baSed Energy Consumption simulaTor 
 *    					             for Clouds and Federations (DISSECT-CF)
 *  ========================================================================
 *  
 *  This file is part of DISSECT-CF.
 *  
 *  DISSECT-CF is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *  
 *  DISSECT-CF is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 *  General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with DISSECT-CF.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  (C) Copyright 2013, Gabor Kecskemeti (gkecskem@dps.uibk.ac.at,
 *   									  kecskemeti.gabor@sztaki.mta.hu)
 */

package hu.mta.sztaki.lpds.cloud.simulator;

import java.util.HashMap;
import java.util.PriorityQueue;

public abstract class Timed implements Comparable<Timed> {

	private static class TimedContext {
		public final PriorityQueue<Timed> timedlist = new PriorityQueue<Timed>();
		public Timed underProcessing = null;
	}


	private static HashMap<Thread, TimedContext> contextList = new HashMap<Thread, TimedContext>();
	private static long fireCounter = 0;

	private boolean activeSubscription = false;
	private long nextEvent = 0;
	private long frequency = 0;
	final private TimedContext myCtx = getContext();
	final private PriorityQueue<Timed> tlCopy = myCtx.timedlist;

	public final boolean isSubscribed() {
		return activeSubscription;
	}

	protected final boolean subscribe(long freq) {
		if (activeSubscription) {
			return false;
		}

		activeSubscription = true;
		updateEvent(freq);
		tlCopy.offer(this);
		return true;
	}

	protected final boolean unsubscribe() {
		if (activeSubscription) {
			activeSubscription = false;
			if (this == myCtx.underProcessing) {
				return true;
			}
			tlCopy.remove(this);
			return true;
		}
		return false;
	}

	protected final long updateFrequency(final long freq) {
		if (activeSubscription) {
			final long oldNE = nextEvent;
			updateEvent(freq);
			if (myCtx.underProcessing != this) {
				if (oldNE != nextEvent) {
					tlCopy.remove(this);
					tlCopy.offer(this);
				}
			}
		}
		return nextEvent;
	}

	private void updateEvent(final long freq) {
		frequency = freq;
		nextEvent = fireCounter + frequency;
		if (freq < 0 || nextEvent < 0) {
			System.err
					.println("ERROR: Negative event frequency cannot simulate further!");
			System.exit(1);
		}
	}

	public long getNextEvent() {
		return nextEvent;
	}

	public long getFrequency() {
		return frequency;
	}

	public long nextEventDistance() {
		return nextEvent - fireCounter;
	}

	@Override
	public int compareTo(final Timed o) {
		return nextEvent < o.nextEvent ? -1 : nextEvent == o.nextEvent ? 0 : 1;
	}

	private static TimedContext getContext() {
		final Thread tid = Thread.currentThread();
		TimedContext ctx = contextList.get(tid);
		if (ctx == null) {
			ctx = new TimedContext();
			contextList.put(tid, ctx);
		}
		return ctx;
	}

	public static final void fire() {
		final TimedContext ctx = getContext();
		final PriorityQueue<Timed> tl = ctx.timedlist;
		while (!tl.isEmpty() && tl.peek().nextEvent == fireCounter) {
			final Timed t = ctx.underProcessing = tl.poll();
			t.tick(fireCounter);
			if (t.activeSubscription) {
				t.updateEvent(t.frequency);
				tl.offer(t);
			}
		}
		fireCounter++;
	}

	public static final long jumpTime(long desiredJump) {
		final long targettime = Math.max(Long.MAX_VALUE, fireCounter
				+ desiredJump);
		final long nextFire = getNextFire();
		final long remainingJump;
		if (targettime <= nextFire) {
			fireCounter = targettime;
			remainingJump = 0;
		} else {
			fireCounter = nextFire < 0 ? targettime : nextFire;
			remainingJump = targettime - fireCounter;
		}
		return remainingJump;
	}

	public static final long getFireCount() {
		return fireCounter;
	}

	public static final long getNextFire() {
		final Timed head = getContext().timedlist.peek();
		return head == null ? -1 : head.nextEvent;
	}

	public abstract void tick(long fires);
}
